"""This module is for the translation mapping data of the
following document:

Document file: int_and_number_comparison_operations.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Int and Number basic comparison operations':
    '# Int と Number クラスの基本的な比較制御',

    'This page explains basic comparison operations of the `Int` and `Number` classes, like the `>=`\\, `<`\\.':  # noqa
    'このページでは`Int`や`Number`クラスの`>=`や`<`などの基本的な比較制御について説明します。',

    '## Common behaviors':
    '## 共通の挙動',

    'Each comparison operation returns a `Boolean` value, not a Python built-in `bool` value:':  # noqa
    '各比較制御はPythonのビルトインの`bool`の値ではなく`Boolean`の値を返却します。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == 10\nassert isinstance(result, ap.Boolean)\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == 10\nassert isinstance(result, ap.Boolean)\n```',  # noqa

    'You can compare the `Int` or `Number` values with the Python built-in values, like the `int` or `float`:':  # noqa
    '`Int`や`Number`の値をPythonビルトインの`int`や`float`などの値と比較することもできます:',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(20)\nresult: ap.Boolean = int_1 == 20\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(20)\nresult: ap.Boolean = int_1 == 20\nassert result\n```',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nnumber_1: ap.Number = ap.Number(10.5)\nresult: ap.Boolean = number_1 == 10.5\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nnumber_1: ap.Number = ap.Number(10.5)\nresult: ap.Boolean = number_1 == 10.5\nassert result\n```',  # noqa

    'Also, the comparison between the `Int` and `Int`\\, `Number` and `Number`\\, `Int` and `Number` are supported:':  # noqa
    '同様に`Int`と`Int`間、`Number`と`Number`間、`Int`と`Number`間の比較などもサポートしています:',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nint_2: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == int_2\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nint_2: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == int_2\nassert result\n```',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nnumber_1: ap.Number = ap.Number(10.5)\nnumber_2: ap.Number = ap.Number(10.5)\nresult: ap.Boolean = number_1 == number_2\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nnumber_1: ap.Number = ap.Number(10.5)\nnumber_2: ap.Number = ap.Number(10.5)\nresult: ap.Boolean = number_1 == number_2\nassert result\n```',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nnumber_1: ap.Number = ap.Number(10)\nresult: ap.Boolean = int_1 == number_1\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nnumber_1: ap.Number = ap.Number(10)\nresult: ap.Boolean = int_1 == number_1\nassert result\n```',  # noqa

    '## Equal comparison operator':
    '## 等値条件の比較のオペレーター',

    'You can use the `==` operator for the equal comparison:':
    '`==`のオペレーターを使って等値条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == 10\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 == 10\nassert result\n```',  # noqa

    '## Not equal comparison operator':
    '## 非等値条件の比較のオペレーター',

    'You can use the `!=` operator for the not equal comparison:':
    '`!=`のオペレーターを使って非等値条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 != 15\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 != 15\nassert result\n```',  # noqa

    '## Less than comparison operator':
    '## 未満条件の比較のオペレーター',

    'You can use the `<` operator for the less than comparison:':
    '`<`のオペレーターを使って未満条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 < 11\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 < 11\nassert result\n```',  # noqa

    '## Less than or equal comparison operator':
    '## 以下条件の比較のオペレーター',

    'You can use the `<=` operator for the less than or equal comparison:':
    '`<=`のオペレーターを使って以下条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 <= 10\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 <= 10\nassert result\n```',  # noqa

    '## Greater than comparison operator':
    '## 超過条件の比較のオペレーター',

    'You can use the `>` operator for the greater than comparison:':
    '`>`のオペレーターを使って超過条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 > 9\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 > 9\nassert result\n```',  # noqa

    '## Greater than or equal comparison operator':
    '## 以上条件の比較のオペレーター',

    'You can use the `>=` operator for the greater than or equal comparison:':  # noqa
    '`>=`のオペレーターを使って以上条件の比較を行うことができます。',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 >= 10\nassert result\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nresult: ap.Boolean = int_1 >= 10\nassert result\n```',  # noqa

}



"""This module is for the translation mapping data of the
following document:

Document file: assertion_basic_behavior.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# JavaScript assertion interface basic behaviors':
    '# JavaScript 上のアサーションのインターフェイスの基本的な挙動',

    'This page explains the JavaScript assertion interface basic behavior.':
    'このページではJavaScript上でのアサーション（テストなどのチェック処理）の各インターフェイスの基本的な挙動について説明します。',

    '## Interface names':
    '## 各インターフェイス名について',

    'Each JavaScript assertion interface has the prefix of the `assert_` (e.g., `assert_equal`, `assert_true`, and so on).':  # noqa
    'JavaScript上の各アサーションのインターフェイスは`assert_`のプレフィックスを持っています（例 : `assert_equal`や`assert_true`など）。',  # noqa

    'These interfaces are positioned in the root package so you can use them, for example, `ap.assert_equal(...)`.':  # noqa
    'これらのインターフェイスはapyscのルートのパッケージに配置されているため、たとえば`ap.assert_equal(...)`といった記述で使うことができます。',  # noqa

    '## Assertion results':
    '## アサーションの結果について',

    'These interfaces display the results on the browser console, as follows:':  # noqa
    'これらのインターフェイスのチェック結果は以下の例のようにブラウザのコンソール上に表示されます:',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=10, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=10, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results/\')\n```',  # noqa

    'This code displays the information message on the browser console, like this (please press the F12 key to see):':  # noqa
    '上記のコード例ではブラウザのコンソール上に以下のような結果の情報が表示されます（ブラウザ上でF12キーを押して確認してください）:',

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 10 right value: 10\n```':  # noqa
    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 10 right value: 10\n```',  # noqa

    'If the assertion fails, then an error message also is displayed on the browser console:':  # noqa
    'もしチェック処理に失敗した場合も同様にブラウザのコンソール上にエラーメッセージが表示されます:',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results_failed/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results_failed/\')\n```',  # noqa

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed:\n...\n```':  # noqa
    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed:\n...\n```',  # noqa

    '## Optional msg argument':
    '## 省略可能なmsg引数について',

    'Each assertion interface has the `msg` optional argument. If you provide this argument, the error message is displayed on the browser console when an assertion fails.':  # noqa
    '各インターフェイスは共通して`msg`という省略可能な引数のとオプションを持っています。もしこの引数に値を指定した場合、チェック処理に失敗した場合にエラーの詳細のメッセージとしてブラウザのコンソール上に表示されます。',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1, msg=\'Values are not equal!\')\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_msg/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1, msg=\'Values are not equal!\')\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_msg/\')\n```',  # noqa

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed: Values are not equal!\n```':  # noqa
    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed: Values are not equal!\n```',  # noqa

}



"""This module is for the geometry-related values' validation
helper interfaces.
"""


from typing import Any


def validate_point_2d_type(
        *, point: Any,
        additional_err_msg: str = '') -> None:
    """
    Validate specified value's type is Point2D.

    Parameters
    ----------
    point : Point2D
        Point2D instance to be checked.
    additional_err_msg : str, optional
        An additional error message to display.

    Raises
    ------
    ValueError
        If a specified value's type is not Point2D.
    """
    import apysc as ap
    if isinstance(point, ap.Point2D):
        return
    if additional_err_msg != '':
        additional_err_msg = f'\n{additional_err_msg}'
    raise ValueError(
        f'Specified value\'s type is not Point2D: {type(point)}'
        f'{additional_err_msg}')



"""Class implementation for the scale_x_from_center interface.
"""

from typing import Dict

from apysc._animation.animation_scale_x_from_center_interface import \
    AnimationScaleXFromCenterInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.number import Number
from apysc._type.revert_interface import RevertInterface
from apysc._validation import arg_validation_decos


class ScaleXFromCenterInterface(
        AnimationScaleXFromCenterInterface, RevertInterface,
        AttrLinkingInterface):

    _scale_x_from_center: Number

    @add_debug_info_setting(
        module_name=__name__, class_name='ScaleXFromCenterInterface')
    def _initialize_scale_x_from_center_if_not_initialized(self) -> None:
        """
        Initialize the `_scale_x_from_center` attribute if
        this instance does not initialize it yet.
        """
        import apysc as ap
        if hasattr(self, '_scale_x_from_center'):
            return
        self._scale_x_from_center = ap.Number(1.0)

        self._append_scale_x_from_center_attr_linking_setting()

    @add_debug_info_setting(
        module_name=__name__, class_name='ScaleXFromCenterInterface')
    def _append_scale_x_from_center_attr_linking_setting(self) -> None:
        """
        Append a scale-x attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._scale_x_from_center,
            attr_name='scale_x_from_center')
        self._append_attr_to_linking_stack(
            attr=self._scale_x_from_center,
            attr_name='scale_x_from_center')

    @property
    @add_debug_info_setting(
        module_name=__name__, class_name='ScaleXFromCenterInterface')
    def scale_x_from_center(self) -> Number:
        """
        Get a scale-x value from the center of this instance.

        Returns
        -------
        scale_x_from_center : ap.Number
            Scale-x value from the center of this instance.

        References
        ----------
        - GraphicsBase scale_x_from_center and scale_y_from_center interfaces
            - https://simon-ritchie.github.io/apysc/en/graphics_base_scale_from_center.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.scale_x_from_center = ap.Number(1.5)
        >>> rectangle.scale_x_from_center
        Number(1.5)
        """
        from apysc._type import value_util
        self._initialize_scale_x_from_center_if_not_initialized()
        return value_util.get_copy(value=self._scale_x_from_center)

    @scale_x_from_center.setter
    @arg_validation_decos.is_apysc_num(arg_position_index=1)
    @add_debug_info_setting(
        module_name=__name__, class_name='ScaleXFromCenterInterface')
    def scale_x_from_center(self, value: Number) -> None:
        """
        Update a scale-x value from the center of this instance.

        Parameters
        ----------
        value : ap.Number
            Scale-x value from the center of this instance.

        References
        ----------
        - GraphicsBase scale_x_from_center and scale_y_from_center interfaces
            - https://simon-ritchie.github.io/apysc/en/graphics_base_scale_from_center.html  # noqa
        """
        import apysc as ap
        self._initialize_scale_x_from_center_if_not_initialized()
        before_value: ap.Number = self._scale_x_from_center
        self._scale_x_from_center = value
        self._append_scale_x_from_center_update_expression(
            before_value=before_value)

        self._append_scale_x_from_center_attr_linking_setting()

    @add_debug_info_setting(
        module_name=__name__, class_name='ScaleXFromCenterInterface')
    def _append_scale_x_from_center_update_expression(
            self, *, before_value: Number) -> None:
        """
        Append the scale-x from the center of this instance
        updating expression.

        Parameters
        ----------
        before_value : ap.Number
            Before updating value.
        """
        import apysc as ap
        from apysc._type import value_util
        before_value_str: str = value_util.get_value_str_for_expression(
            value=before_value)
        after_value_str: str = value_util.get_value_str_for_expression(
            value=self._scale_x_from_center)
        expression: str = (
            f'{self.variable_name}.scale(1 / {before_value_str}, 1);'
            f'\n{self.variable_name}.scale({after_value_str}, 1);'
            f'\n{before_value_str} = {after_value_str};'
        )
        ap.append_js_expression(expression=expression)

    _scale_x_from_center_snapshots: Dict[str, float]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_scale_x_from_center_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_scale_x_from_center_snapshots',
            value=self._scale_x_from_center._value,
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._scale_x_from_center._value = \
            self._scale_x_from_center_snapshots[snapshot_name]



"""This module is for the handler interfaces'
validation implementations.
"""

import inspect
from inspect import Signature
from typing import Any
from typing import Callable
from typing import List


def validate_options_type(
        *, options: Any,
        additional_err_msg: str = '') -> None:
    """
    Validate a specified options type.

    Parameters
    ----------
    options : Any
        Target options value.
    additional_err_msg : str, optional
        An additional error message to display.

    Raises
    ------
    TypeError
        If a specified options type is not the dictionary or None.
    """
    if options is None:
        return
    if isinstance(options, dict):
        return
    if additional_err_msg != '':
        additional_err_msg = f'\n{additional_err_msg}'
    raise TypeError(
        f"Handler's options argument must be a dictionary: {type(options)}"
        f'\n{options}{additional_err_msg}')


def validate_handler_args_num(
        *, handler: Callable,
        additional_err_msg: str = '') -> None:
    """
    Validate specified handler's arguments number.

    Parameters
    ----------
    handler : Callable
        A target handler to validate.
    additional_err_msg : str, optional
        An additional error message to display.

    Raises
    ------
    ValueError
        - If handler's arguments number is not 2.
    TypeError
        - If a specified handler is not callable.
    """
    if not callable(handler):
        raise TypeError(
            'A specified handler\'s argument is not callable: '
            f'{type(handler)}'
        )
    signature: Signature = inspect.signature(obj=handler)
    args_num: int = 0
    skipping_arg_names: List[str] = ['*', '**', 'self', 'cls']
    arg_names: List[str] = []
    for parameter in signature.parameters:
        if parameter in skipping_arg_names:
            continue
        args_num += 1
        arg_names.append(parameter)
    if args_num != 2:
        if additional_err_msg != '':
            additional_err_msg = f'\n{additional_err_msg}'
        raise ValueError(
            'A specified handler\'s arguments number must be 2 '
            f'(actual: {args_num})'
            f'\nTarget argument names: {arg_names}'
            '\n\nThe first argument becomes event instance and the second '
            'one becomes the handler\'s option parameters.'
            f'{additional_err_msg}')



"""This module is for the translation mapping data of the
following document:

Document file: display_object.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# DisplayObject class':
    '# DisplayObject クラス',

    'This page explains the `DisplayObject` class.':
    'このページでは`DisplayObject`クラスについて説明します。',

    '## What is the DisplayObject?':
    '## DisplayObject クラスの概要',

    'The `DisplayObject` is the apysc base class for each display class, such as  `Sprite`\\, `Rectangle`\\, `Circle`\\, or something else.':  # noqa
    '`DisplayObject`クラスは`Sprite`や`Rectangle`などのapyscの表示オブジェクトの基底クラスとなります。',

    'You can verify the `DisplayObject` inheritance with each instance by the `isinstance` function.':  # noqa
    '各インスタンスが`DisplayObject`を継承していることを`isinstance`関数を使って確認することができます。',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=200,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\ncircle: ap.Circle = sprite.graphics.draw_circle(x=100, y=100, radius=100)\n\n# Verify each instance type.\nassert isinstance(sprite, ap.DisplayObject)\nassert isinstance(circle, ap.DisplayObject)\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=200,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\ncircle: ap.Circle = sprite.graphics.draw_circle(x=100, y=100, radius=100)\n\n# Verify each instance type.\nassert isinstance(sprite, ap.DisplayObject)\nassert isinstance(circle, ap.DisplayObject)\n```',  # noqa

    'The apysc uses this class for the basic interfaces or the creating the new display object with the `DisplayObject` inheritance.':  # noqa
    'apyscはこのクラスを基本的な共通のインターフェイスで使用したり、もしくは別の`DisplayObject`を継承したインスタンスの作成処理などで使用しています。',  # noqa

    'The `DisplayObject` class has the basic interfaces, like `x`\\, `y`\\, `visible`\\, each mouse event binding, or others. The following page explains these interfaces one by one.':  # noqa
    '`DisplayObject`クラスは`x`や`y`、`visible`属性、各マウスイベントの設定などの基本的なインターフェイスを持っています。以下のページではそれぞれのインターフェイスについて1つ1つ詳しく触れています。',  # noqa

    '## See also':
    '## 関連資料',

    '- [DisplayObject class x and y interfaces](display_object_x_and_y.md)':
    '- [DisplayObject クラスの x と y インターフェイス](jp_display_object_x_and_y.md)',

    '- [DisplayObject class parent interfaces](display_object_parent.md)':
    '- [DisplayObjectクラス parent （親要素属性）のインターフェイス](jp_display_object_parent.md)',  # noqa

    '- [DisplayObject class visible interface](display_object_visible.md)':
    '- [DisplayObject クラスの visible (表示・非表示) のインターフェイス](jp_display_object_visible.md)',  # noqa

    '- [DisplayObject class mouse event binding interfaces](display_object_mouse_event.md)':  # noqa
    '- [DisplayObject クラスのマウスイベント設定の各インターフェイス](jp_display_object_mouse_event.md)',  # noqa

}



"""This module is for the translation mapping data of the
following document:

Document file: array_clear.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class clear interface':
    '# Array クラスの clear インターフェイス',

    'This page explains the `Array` class `clear` method interface.':
    'このページでは`Array`クラスの`clear`メソッドのインターフェイスについて説明します。',

    '## What interface is this?':
    '## インターフェイス概要',

    'The `clear` method clears an Array\'s value. This interface makes an Array\'s value empty.':  # noqa
    '`clear`メソッドは配列の値を空にします。',

    '## Basic usage':
    '## 基本的な使い方',

    'The `clear` method requires no arguments.':
    '`clear`メソッドは呼び出しに特に引数を必要としません。',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array = ap.Array([10, 20, 30])\narr.clear()\nassert arr == []\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array = ap.Array([10, 20, 30])\narr.clear()\nassert arr == []\n```',  # noqa

}



"""Class implementation for the center y-coordinate interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_cy_interface import AnimationCyInterface
from apysc._display.y_interface_base import YInterfaceBase
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._validation import arg_validation_decos


class CyInterface(
        YInterfaceBase, AnimationCyInterface, RevertInterface,
        AttrLinkingInterface):

    def _initialize_y_if_not_initialized(self) -> None:
        """
        Initialize _y attribute if this interface does not
        initialize it yet.
        """
        if hasattr(self, '_y'):
            return
        self._y = Int(0)

        self._append_y_attr_linking_setting()

    @add_debug_info_setting(
        module_name=__name__, class_name='CyInterface')
    def _append_y_attr_linking_setting(self) -> None:
        """
        Append y attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._y, attr_name='y')
        self._append_attr_to_linking_stack(attr=self._y, attr_name='y')

    @property
    @add_debug_info_setting(
        module_name=__name__, class_name='CyInterface')
    def y(self) -> Int:
        """
        Get a center y-coordinate.

        Returns
        -------
        y : Int
            Center y-coordinate.

        References
        ----------
        - Display object x and y interfaces document
            - https://simon-ritchie.github.io/apysc/en/display_object_x_and_y.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> circle: ap.Circle = sprite.graphics.draw_circle(
        ...     x=100, y=100, radius=50)
        >>> circle.y = ap.Int(120)
        >>> circle.y
        Int(120)
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_y_if_not_initialized()
        y: ap.Int = value_util.get_copy(value=self._y)
        return y

    @y.setter
    @arg_validation_decos.is_apysc_num(arg_position_index=1)
    @add_debug_info_setting(
        module_name=__name__, class_name='CyInterface')
    def y(self, value: Int) -> None:
        """
        Update a center y-coordinate.

        Parameters
        ----------
        value : Int
            Center y-coordinate value.

        References
        ----------
        - Display object x and y interfaces document
            - https://simon-ritchie.github.io/apysc/en/display_object_x_and_y.html  # noqa
        """
        self._y = value
        self._y._append_incremental_calc_substitution_expression()
        self._append_y_update_expression()

        self._append_y_attr_linking_setting()

    @add_debug_info_setting(
        module_name=__name__, class_name='CyInterface')
    def _append_y_update_expression(self) -> None:
        """
        Append y position updating expression.
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_y_if_not_initialized()
        value_str: str = value_util.get_value_str_for_expression(
            value=self._y)
        expression: str = (
            f'{self.variable_name}.cy({value_str});'
        )
        ap.append_js_expression(expression=expression)

    def _update_y_and_skip_appending_exp(
            self, *, y: Union[int, Int]) -> None:
        """
        Update y-coordinate and skip appending an expression.

        Parameters
        ----------
        y : int or Int
            Y-coordinate value.
        """
        if isinstance(y, Int):
            y_: Int = y
        else:
            y_ = Int(y)
        self._y = y_

    _y_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_y_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_y_snapshots',
            value=int(self._y._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._y._value = self._y_snapshots[snapshot_name]



"""Implementations of Polyline class.
"""


from typing import List
from typing import Optional as Op
from typing import Union

from apysc._display import graphics
from apysc._display.append_line_point_interface import AppendLinePointInterface
from apysc._display.child_interface import ChildInterface
from apysc._display.graphics_base import GraphicsBase
from apysc._display.line_caps import LineCaps
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_joints import LineJoints
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.x_interface import XInterface
from apysc._display.y_interface import YInterface
from apysc._geom.point2d import Point2D
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String
from apysc._validation import arg_validation_decos


class Polyline(
        XInterface,
        YInterface,
        GraphicsBase,
        AppendLinePointInterface):
    """
    The polyline vector graphics class.

    References
    ----------
    - Graphics move_to and line_to interfaces document
        - https://simon-ritchie.github.io/apysc/en/graphics_move_to_and_line_to.html  # noqa

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.line_style(color='#fff', thickness=5)
    >>> _ = sprite.graphics.move_to(x=50, y=50)
    >>> polyline: ap.Polyline = sprite.graphics.line_to(x=150, y=50)
    >>> polyline.line_color
    String('#ffffff')

    >>> polyline.line_thickness
    Int(5)
    """

    # self
    @arg_validation_decos.multiple_line_settings_are_not_set(
        arg_position_index=0)
    # points
    @arg_validation_decos.is_point_2ds(arg_position_index=1)
    # fill_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=2)
    # fill_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=3)
    # line_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=4)
    # line_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=5)
    # line_thickness
    @arg_validation_decos.is_integer(arg_position_index=6)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=6)
    # line_cap
    @arg_validation_decos.is_line_cap(
        arg_position_index=7, optional=True)
    # line_joints
    @arg_validation_decos.is_line_joints(
        arg_position_index=8, optional=True)
    # line_dot_setting
    @arg_validation_decos.is_line_dot_setting(arg_position_index=9)
    # line_dash_setting
    @arg_validation_decos.is_line_dash_setting(arg_position_index=10)
    # line_round_dot_setting
    @arg_validation_decos.is_line_round_dot_setting(arg_position_index=11)
    # line_dash_dot_setting
    @arg_validation_decos.is_line_dash_dot_setting(arg_position_index=12)
    # parent
    @arg_validation_decos.is_display_object_container(
        arg_position_index=13, optional=True)
    @add_debug_info_setting(
        module_name=__name__, class_name='Polyline')
    def __init__(
            self, *,
            points: Union[Array[Point2D], List[Point2D]],
            fill_color: Union[str, String] = '',
            fill_alpha: Union[float, Number] = 1.0,
            line_color: Union[str, String] = '',
            line_alpha: Union[float, Number] = 1.0,
            line_thickness: Union[int, Int] = 1,
            line_cap: Op[Union[String, LineCaps]] = None,
            line_joints: Op[Union[String, LineJoints]] = None,
            line_dot_setting: Op[LineDotSetting] = None,
            line_dash_setting: Op[LineDashSetting] = None,
            line_round_dot_setting: Op[LineRoundDotSetting] = None,
            line_dash_dot_setting: Op[LineDashDotSetting] = None,
            parent: Op[ChildInterface] = None) -> None:
        """
        Create a polyline vector graphic.

        Parameters
        ----------
        points : Array of Point2D or list of Point2D
            List of line points.
        fill_color : str or String, default ''
            A fill-color to set.
        fill_alpha : float or Number, default 1.0
            A fill-alpha to set.
        line_color : str or String, default ''
            A line-color to set.
        line_alpha : float or Number, default 1.0
            A line-alpha to set.
        line_thickness : int or Int, default 1
            A line-thickness (line-width) to set.
        line_cap : String or LineCaps or None, default None
            A line-cap setting to set.
        line_joints : String or LineJoints or None, default None
            A line-joints setting to set.
        line_dot_setting : LineDotSetting or None, default None
            A dot setting to set.
        line_dash_setting : LineDashSetting or None, default None
            A dash setting to set.
        line_round_dot_setting : LineRoundDotSetting or None, default None
            A round-dot setting to set.
        line_dash_dot_setting : LineDashDotSetting or None, default None
            A dash dot (1-dot chain) setting to set.
        parent : ChildInterface or None, default None
            A parent instance to add this instance.
            If a specified value is None, this interface uses
            a stage instance.
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        if isinstance(points, list):
            points = Array(points)
        variable_name: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.POLYLINE)
        self.variable_name = variable_name
        self.points = points
        self._set_initial_basic_values(
            fill_color=fill_color, fill_alpha=fill_alpha,
            line_color=line_color, line_thickness=line_thickness,
            line_alpha=line_alpha, line_cap=line_cap, line_joints=line_joints)
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            line_dot_setting=line_dot_setting,
            line_dash_setting=line_dash_setting,
            line_round_dot_setting=line_round_dot_setting,
            line_dash_dot_setting=line_dash_dot_setting)
        super(Polyline, self).__init__(
            parent=parent, variable_name=variable_name)

    @classmethod
    def _create_with_graphics(
            cls,
            *,
            graphics: 'graphics.Graphics',
            points: Union[Array[Point2D], List[Point2D]]) -> 'Polyline':
        """
        Create a polyline instance with the instance of
        specified graphics.

        Parameters
        ----------
        graphics : Graphics
            Graphics instance to link this instance.
        points : Array of Point2D or list of Point2D
            List of line points._description_

        Returns
        -------
        polyline : Polyline
            A created polyline instance.
        """
        polyline: Polyline = Polyline(
            points=points,
            fill_color=graphics._fill_color,
            fill_alpha=graphics._fill_alpha,
            line_color=graphics._line_color,
            line_alpha=graphics._line_alpha,
            line_thickness=graphics._line_thickness,
            line_cap=graphics._line_cap,
            line_joints=graphics._line_joints,
            line_dot_setting=graphics._line_dot_setting,
            line_dash_setting=graphics._line_dash_setting,
            line_round_dot_setting=graphics._line_round_dot_setting,
            line_dash_dot_setting=graphics._line_dash_dot_setting,
            parent=graphics)
        return polyline

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Polyline('<variable_name>')`).
        """
        repr_str: str = f"Polyline('{self.variable_name}')"
        return repr_str

    @add_debug_info_setting(
        module_name=__name__, class_name='Polyline')
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        from apysc._display.stage import get_stage_variable_name
        stage_variable_name: str = get_stage_variable_name()
        points_var_name: str
        points_expression: str
        points_var_name, points_expression = \
            self._make_2dim_points_expression()
        expression: str = (
            f'{points_expression}'
            f'\nvar {self.variable_name} = {stage_variable_name}'
            f'\n  .polyline({points_var_name})'
            '\n  .attr({'
        )
        expression = self._append_basic_vals_expression(
            expression=expression, indent_num=2)
        expression += '\n  });'
        ap.append_js_expression(expression=expression)
        self._points_var_name = points_var_name



"""This module is for the translation mapping data of the
following document:

Document file: draw_interfaces_abstract.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Draw interfaces abstract':
    '# 描画の各インターフェイスの概要',

    'This page would explain the drawing interfaces abstract.':
    'このページでは描画の各インターフェイスについて説明します。',

    '## What apysc can do in its drawing interfaces':
    '## apyscの描画の各インターフェイスでできること',

    '- You can set the fill-color, fill-alpha (opacity), line-color, line-alpha, line-thickness values and draw each SVG graphic with these.':  # noqa
    '- これらのインターフェイスを使って塗りの色、塗りの透明度、線の色、線の透明度、線幅などを設定することができます。',

    '- Supported graphics: The rectangle, circle, ellipse, polygon, line, polyline, and path.':  # noqa
    '- 四角や丸、楕円、多角形、線、折れ線、パスなどの描画をサポートしています。',

    '## Fill settings':
    '## 塗りの設定',

    'The `begin_fill` interface sets the fill-color and fill-alpha (opacity).':  # noqa
    '`begin_fill`のインターフェイスは塗りの色と塗りの透明度を設定します。',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_begin_fill/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_begin_fill/\')\n```',  # noqa

    'For more details, please see:':
    '詳細は以下をご確認ください:',

    '- [Graphics class begin_fill interface](graphics_begin_fill.md).':
    '- [Graphics クラスの begin_fill (塗りの設定)のインターフェイス](jp_graphics_begin_fill.md)',  # noqa

    '## Line style settings':
    '## 線のスタイル設定',

    'The `line_style` interface sets the line-color, line-alpha (opacity), line-thickness.':  # noqa
    '`line_style`インターフェイスは線の色と線の透明度、線幅などを設定することができます。',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=100,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color=\'#fff\', thickness=5, alpha=0.5)\nsprite.graphics.draw_line(x_start=50, y_start=50, x_end=150, y_end=50)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_line_style/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=100,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color=\'#fff\', thickness=5, alpha=0.5)\nsprite.graphics.draw_line(x_start=50, y_start=50, x_end=150, y_end=50)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_line_style/\')\n```',  # noqa

    'For more details, please see:':
    '詳細は以下をご確認ください:',

    '- [Graphics class line_style interface](graphics_line_style.md).':
    '- [Graphics クラスの line_style (線のスタイル設定)のインターフェイス](jp_graphics_line_style.md)',  # noqa

    '## Each drawing interface':
    '## 描画の各インターフェイス',

    'Each drawing interface has the `draw_` prefix draw SVG graphics (e.g., draw_rect, draw_circle).':  # noqa
    '各描画のインターフェイスは`draw_`のプレフィックスを持っており、SVGのグラフィックを描画します（例 : draw_rect や draw_circle など）。',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_circle(x=175, y=75, radius=25)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_each_drawing_interface/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_circle(x=175, y=75, radius=25)\n\nap.save_overall_html(\n    dest_dir_path=\'draw_interfaces_abstract_each_drawing_interface/\')\n```',  # noqa

    'For more details, please see the following:':
    '詳細については以下をご確認ください:',

    '- [Graphics class draw_rect interface](graphics_draw_rect.md)':
    '- [Graphics クラスの draw_rect (四角の描画)のインターフェイス](jp_graphics_draw_rect.md)',  # noqa

    '- [Graphics class draw_round_rect interface](graphics_draw_round_rect.md)':  # noqa
    '- [Graphics クラスの draw_round_rect (角丸の四角の描画)のインターフェイス](jp_graphics_draw_round_rect.md)',  # noqa

    '- [Graphics class draw_circle interface](graphics_draw_circle.md)':
    '- [Graphics クラスの draw_circle (円の描画)のインターフェイス](jp_graphics_draw_circle.md)',  # noqa

    '- [Graphics class draw_ellipse interface](graphics_draw_ellipse.md)':
    '- [Graphics クラスの draw_ellipse (楕円描画) のインターフェイス](jp_graphics_draw_ellipse.md)',  # noqa

    '- [Graphics class move_to and line_to interfaces](graphics_move_to_and_line_to.md)':  # noqa
    '- [Graphics クラスの move_to (線の描画位置の変更)と line_to (指定座標への線の描画)のインターフェイス](jp_graphics_move_to_and_line_to.md)',  # noqa

    '- [Graphics class draw_line interface](graphics_draw_line.md)':
    '- [Graphics クラスの draw_line (線の描画)のインターフェイス](jp_graphics_draw_line.md)',

    '- [Graphics class draw_dotted_line interface](graphics_draw_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_dotted_line (点線の描画)のインターフェイス](jp_graphics_draw_dotted_line.md)',  # noqa

    '- [Graphics class draw_dashed_line interface](graphics_draw_dashed_line.md)':  # noqa
    '- [Graphics クラスの draw_dashed_line (破線の描画)のインターフェイス](jp_graphics_draw_dashed_line.md)',  # noqa

    '- [Graphics class draw_round_dotted_line interface](graphics_draw_round_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_round_dotted_line (点線(丸)の描画)のインターフェイス](jp_graphics_draw_round_dotted_line.md)',  # noqa

    '- [Graphics class draw_dash_dotted_line interface](graphics_draw_dash_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_dash_dotted_line (一点鎖線の描画)のインターフェイス](jp_graphics_draw_dash_dotted_line.md)',  # noqa

    '- [Graphics class draw_polygon interface](graphics_draw_polygon.md)':
    '- [Graphics クラスの draw_polygon (多角形描画)のインターフェイス](jp_graphics_draw_polygon.md)',  # noqa

    '## See also':
    '## 関連資料',

    '- [Graphics class](graphics.md)':
    '- [Graphics クラス](jp_graphics.md)',

    '- [Graphics class fill_color interface](graphics_fill_color.md)':
    '- [Graphics クラスの fill_color (塗り設定)のインターフェイス](jp_graphics_fill_color.md)',  # noqa

    '- [Graphics class fill_alpha interface](graphics_fill_alpha.md)':
    '- [Graphics クラスの fill_alpha (塗りの透明度設定)のインターフェイス](jp_graphics_fill_alpha.md)',  # noqa

    '- [Graphics class line_color interface](graphics_line_color.md)':
    '- [Graphics クラスの line_color (線の色設定)のインターフェイス](jp_graphics_line_color.md)',  # noqa

    '- [Graphics class line_alpha interface](graphics_line_alpha.md)':
    '- [Graphics クラスの line_color (線の透明度設定)のインターフェイス](jp_graphics_line_alpha.md)',  # noqa

    '- [Graphics class line_thickness interface](graphics_line_thickness.md)':  # noqa
    '- [Graphics クラスの line_color (線幅設定)のインターフェイス](jp_graphics_line_thickness.md)',  # noqa

    '- [Graphics class line_dot_setting interface](graphics_line_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_dot_setting (点線設定)のインターフェイス](jp_graphics_line_dot_setting.md)',  # noqa

    '- [Graphics class line_dash_setting interface](graphics_line_dash_setting.md)':  # noqa
    '- [Graphics クラスの line_dash_setting (破線設定)のインターフェイス](jp_graphics_line_dash_setting.md)',  # noqa

    '- [Graphics class line_round_dot_setting interface](graphics_line_round_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_round_dot_setting (点線(丸)設定)のインターフェイス](jp_graphics_line_round_dot_setting.md)',  # noqa

    '- [Graphics class line_dash_dot_setting interface](graphics_line_dash_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_dash_dot_setting (一点鎖線設定)のインターフェイス](jp_graphics_line_dash_dot_setting.md)',  # noqa

}



"""Implementations of the Path class.
"""

from typing import List
from typing import Optional as Op
from typing import Union

from apysc._display import graphics
from apysc._display.child_interface import ChildInterface
from apysc._display.graphics_base import GraphicsBase
from apysc._display.line_caps import LineCaps
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_joints import LineJoints
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.x_interface import XInterface
from apysc._display.y_interface import YInterface
from apysc._geom.path_data_base import PathDataBase
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String
from apysc._validation import arg_validation_decos


class Path(
        XInterface,
        YInterface,
        GraphicsBase):
    """
    The path vector graphics class.

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.line_style(color='#fff', thickness=3)
    >>> path: ap.Path = sprite.graphics.draw_path(
    ...     path_data_list=[
    ...         ap.PathMoveTo(x=0, y=50),
    ...         ap.PathBezier2D(
    ...             control_x=50, control_y=0,
    ...             dest_x=100, dest_y=50),
    ...     ])
    """

    _path_data_list: List[PathDataBase]

    # self
    @arg_validation_decos.multiple_line_settings_are_not_set(
        arg_position_index=0)
    # path_data_list
    @arg_validation_decos.is_path_data_list(arg_position_index=1)
    # fill_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=2)
    # fill_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=3)
    # line_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=4)
    # line_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=5)
    # line_thickness
    @arg_validation_decos.is_integer(arg_position_index=6)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=6)
    # line_cap
    @arg_validation_decos.is_line_cap(
        arg_position_index=7, optional=True)
    # line_joints
    @arg_validation_decos.is_line_joints(
        arg_position_index=8, optional=True)
    # line_dot_setting
    @arg_validation_decos.is_line_dot_setting(arg_position_index=9)
    # line_dash_setting
    @arg_validation_decos.is_line_dash_setting(arg_position_index=10)
    # line_round_dot_setting
    @arg_validation_decos.is_line_round_dot_setting(arg_position_index=11)
    # line_dash_dot_setting
    @arg_validation_decos.is_line_dash_dot_setting(arg_position_index=12)
    # parent
    @arg_validation_decos.is_display_object_container(
        arg_position_index=13, optional=True)
    @add_debug_info_setting(
        module_name=__name__, class_name='Path')
    def __init__(
            self,
            *,
            path_data_list: List[PathDataBase],
            fill_color: Union[str, String] = '',
            fill_alpha: Union[float, Number] = 1.0,
            line_color: Union[str, String] = '',
            line_alpha: Union[float, Number] = 1.0,
            line_thickness: Union[int, Int] = 1,
            line_cap: Op[Union[String, LineCaps]] = None,
            line_joints: Op[Union[String, LineJoints]] = None,
            line_dot_setting: Op[LineDotSetting] = None,
            line_dash_setting: Op[LineDashSetting] = None,
            line_round_dot_setting: Op[LineRoundDotSetting] = None,
            line_dash_dot_setting: Op[LineDashDotSetting] = None,
            parent: Op[ChildInterface] = None) -> None:
        """
        Create a path vector graphic.

        Parameters
        ----------
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.
        fill_color : str or String, default ''
            A fill-color to set.
        fill_alpha : float or Number, default 1.0
            A fill-alpha to set.
        line_color : str or String, default ''
            A line-color to set.
        line_alpha : float or Number, default 1.0
            A line-alpha to set.
        line_thickness : int or Int, default 1
            A line-thickness (line-width) to set.
        line_cap : String or LineCaps or None, default None
            A line-cap setting to set.
        line_joints : String or LineJoints or None, default None
            A line-joints setting to set.
        line_dot_setting : LineDotSetting or None, default None
            A dot setting to set.
        line_dash_setting : LineDashSetting or None, default None
            A dash setting to set.
        line_round_dot_setting : LineRoundDotSetting or None, default None
            A round-dot setting to set.
        line_dash_dot_setting : LineDashDotSetting or None, default None
            A dash dot (1-dot chain) setting to set.
        parent : ChildInterface or None, default None
            A parent instance to add this instance.
            If a specified value is None, this interface uses
            a stage instance.
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        variable_name: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.PATH)
        self.variable_name = variable_name
        self._set_initial_basic_values(
            fill_color=fill_color, fill_alpha=fill_alpha,
            line_color=line_color, line_thickness=line_thickness,
            line_alpha=line_alpha, line_cap=line_cap, line_joints=line_joints)
        self._path_data_list = path_data_list
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            line_dot_setting=line_dot_setting,
            line_dash_setting=line_dash_setting,
            line_round_dot_setting=line_round_dot_setting,
            line_dash_dot_setting=line_dash_dot_setting)
        super(Path, self).__init__(
            parent=parent, variable_name=variable_name)

    @classmethod
    def _create_with_graphics(
            cls,
            *,
            graphics: 'graphics.Graphics',
            path_data_list: List[PathDataBase]) -> 'Path':
        """
        Create a path instance with the instance of
        specified graphics.

        Parameters
        ----------
        graphics : Graphics
            Graphics instance to link this instance.
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.

        Returns
        -------
        path : Path
            A created path instance.
        """
        path: Path = Path(
            path_data_list=path_data_list,
            fill_color=graphics._fill_color,
            fill_alpha=graphics._fill_alpha,
            line_color=graphics._line_color,
            line_alpha=graphics._line_alpha,
            line_thickness=graphics._line_thickness,
            line_cap=graphics._line_cap,
            line_joints=graphics._line_joints,
            line_dot_setting=graphics._line_dot_setting,
            line_dash_setting=graphics._line_dash_setting,
            line_round_dot_setting=graphics._line_round_dot_setting,
            line_dash_dot_setting=graphics._line_dash_dot_setting,
            parent=graphics)
        return path

    @add_debug_info_setting(
        module_name=__name__, class_name='Path')
    def _append_constructor_expression(self) -> None:
        """
        Append a constructor expression.
        """
        import apysc as ap
        from apysc._display.stage import get_stage_variable_name
        from apysc._geom.path_data_util import make_paths_expression_from_list
        from apysc._string import indent_util
        stage_variable_name: str = get_stage_variable_name()
        path_data_expression: str = make_paths_expression_from_list(
            path_data_list=self._path_data_list)
        INDENT_NUM: int = 2
        expression: str = (
            f'var {self.variable_name} = {stage_variable_name}'
            f'\n  .path({path_data_expression})'
            '\n  .attr({'
        )
        expression = self._append_basic_vals_expression(
            expression=expression,
            indent_num=INDENT_NUM)
        spaces: str = indent_util.make_spaces_for_html(
            indent_num=INDENT_NUM)
        if self._fill_color._value == '':
            expression += f'\n{spaces}fill: "transparent",'
        expression += '\n  });'
        ap.append_js_expression(expression=expression)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Path('<variable_name>')`).
        """
        repr_str: str = f"Path('{self.variable_name}')"
        return repr_str



"""This module is for the translation mapping data of the
following document:

Document file: index.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# apysc documentation':
    '# apysc ドキュメント',

    'Welcome to apysc documentation! apysc is the Python front-end library (currently developing and only works partially).':  # noqa
    'apyscのドキュメントページへようこそ。apyscはPythonのフロントエンド用のライブラリです（現在開発中であり部分的にのみ動作します）。',  # noqa

    '## Project links':
    '## プロジェクトの関連リンク',

    '- [GitHub](https://github.com/simon-ritchie/apysc)':
    '- [GitHub](https://github.com/simon-ritchie/apysc)',

    '  - Please leave a ⭐️star⭐️ if you favor the apysc library or have high hopes for the library\'s future!':  # noqa
    '  - もしもapyscライブラリを気にいったり、もしくはライブラリの将来に期待が出来そうと感じられたらリポジトリにスターを付けていただけますと幸いです。',  # noqa

    '- [PyPI](https://pypi.org/project/apysc/)':
    '- [PyPI](https://pypi.org/project/apysc/)',

    'Other languages documentations: | English | [Japanese (日本語)](https://simon-ritchie.github.io/apysc/jp/jp_index.html) |':  # noqa
    '他の言語のドキュメント: | [英語 (English)](https://simon-ritchie.github.io/apysc/en/index.html) | 日本語 |',  # noqa

    '## Quick start guide':
    '## クイックスタートガイド',

    '### Table of contents':
    '### Table of contents',

    '- [What apysc can do in its current implementation](what_apysc_can_do.md)':  # noqa
    '- [apyscが現在の実装で出来ることの概要](jp_what_apysc_can_do.md)',

    '- [Quick start guide](quick_start.md)':
    '- [クイックスタートガイド](jp_quick_start.md)',

    '- [import conventions](import_conventions.md)':
    '- [import の慣習](jp_import_conventions.md)',

    '- [Recommended type annotation checker settings](recommended_type_checker_settings.md)':  # noqa
    '- [推奨される型アノテーションのチェック設定](jp_recommended_type_checker_settings.md)',

    '---':
    '---',

    '## Container classes':
    '## コンテナーの各クラス',

    'The `Stage` is the apysc overall drawing area container, and the `Sprite` is the container class.':  # noqa
    '`Stage`はapyscの描画領域全体のコンテナーとなり、`Sprite`は通常の各インスタンスのコンテナーのクラスとなります。',

    '### Table of contents':
    '### Table of contents',

    '- [Stage class](stage.md)':
    '- [Stage クラス](jp_stage.md)',

    '- [Sprite class](sprite.md)':
    '- [Sprite クラス](jp_sprite.md)',

    '---':
    '---',

    '## Exporting':
    '## 出力処理',

    'The HTML and JavaScript exporting interfaces.':
    '以下はHTMLとJavaScriptの出力処理関係の各インターフェイスです。',

    '### Table of contents':
    '### Table of contents',

    '- [save_overall_html interface](save_overall_html.md)':
    '- [save_overall_html インターフェイス](jp_save_overall_html.md)',

    '- [display_on_jupyter interface](display_on_jupyter.md)':
    '- [display_on_jupyter インターフェイス](jp_display_on_jupyter.md)',

    '- [display_on_colaboratory interface](display_on_colaboratory.md)':
    '- [display_on_colaboratory インターフェイス](jp_display_on_colaboratory.md)',

    '- [append_js_expression interface](append_js_expression.md)':
    '- [append_js_expression インターフェイス](jp_append_js_expression.md)',

    '---':
    '---',

    '## Child-related interfaces':
    '## 子要素関係の各インターフェイス',

    'The parent class, such as the `Sprite` or `Stage`, has the following interfaces:':  # noqa
    '`Sprite`や`Stage`などの親となれる各クラスは以下のインターフェイスを持っています:',

    '### Table of contents':
    '### Table of contents',

    '- [add_child and remove_child interfaces](add_child_and_remove_child.md)':  # noqa
    '- [add_child （子の追加）と remove_child （子の削除）のインターフェイス](jp_add_child_and_remove_child.md)',  # noqa

    '- [remove_children interface](remove_children.md)':
    '- [remove_children インターフェイス](jp_remove_children.md)',

    '- [contains interface](contains.md)':
    '- [contains インターフェイス](jp_contains.md)',

    '- [num_children interface](num_children.md)':
    '- [num_children （子の件数属性）のインターフェイス](jp_num_children.md)',

    '- [get_child_at interface](get_child_at.md)':
    '- [get_child_at （特定位置の子の取得処理）のインターフェイス](jp_get_child_at.md)',

    '---':
    '---',

    '## apysc basic data classes':
    '## apyscの基本的な各データクラス',

    '### Table of contents':
    '### Table of contents',

    '- [Why the apysc library does not use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '- [なぜapyscではPythonのビルトインのデータの型を使用していないのか](jp_why_apysc_doesnt_use_python_builtin_data_type.md)',  # noqa

    '- [Funcdamental data classes common value interface](fundamental_data_classes_value_interface.md)':  # noqa
    '- [基本的なデータクラスの共通の value インターフェイス](jp_fundamental_data_classes_value_interface.md)',  # noqa

    '- [Int and Number classes](int_and_number.md)':
    '- [Int と Number の各クラス](jp_int_and_number.md)',

    '- [Int and Number classes common arithmetic operations](int_and_number_arithmetic_operations.md)':  # noqa
    '- [Int と Number クラスの共通の計算制御](jp_int_and_number_arithmetic_operations.md)',  # noqa

    '- [Int and Number classes common comparison operations](int_and_number_comparison_operations.md)':  # noqa
    '- [Int と Number クラスの共通の比較制御](jp_int_and_number_comparison_operations.md)',  # noqa

    '- [String class](string.md)':
    '- [String クラス](jp_string.md)',

    '- [String class comparison operations](string_comparison_operations.md)':  # noqa
    '- [String クラスの比較制御](jp_string_comparison_operations.md)',

    '- [String class addition and multiplication operations](string_addition_and_multiplication.md)':  # noqa
    '- [String クラスの加算と乗算の制御](jp_string_addition_and_multiplication.md)',

    '- [Boolean class](boolean.md)':
    '- [Boolean クラス](jp_boolean.md)',

    '- [Array class](array.md)':
    '- [Array クラス](jp_array.md)',

    '- [Array class append and push interfaces](array_append_and_push.md)':
    '- [Array クラスの append と push のインターフェイス](jp_array_append_and_push.md)',

    '- [Array class extend and concat interfaces](array_extend_and_concat.md)':  # noqa
    '- [Array クラスの extend と concat のインターフェイス](jp_array_extend_and_concat.md)',  # noqa

    '- [Array class insert and insert_at interfaces](array_insert_and_insert_at.md)':  # noqa
    '- [Array クラスの insert と insert_at のインターフェイス](jp_array_insert_and_insert_at.md)',  # noqa

    '- [Array class pop interface](array_pop.md)':
    '- [Array クラスの pop インターフェイス](jp_array_pop.md)',

    '- [Array class remove and remove_at interfaces](array_remove_and_remove_at.md)':  # noqa
    '- [Array クラスの remove と remove_at のインターフェイス](jp_array_remove_and_remove_at.md)',  # noqa

    '- [Array class sort interface](array_sort.md)':
    '- [Array クラスの sort インターフェイス](jp_array_sort.md)',

    '- [Array class reverse interface](array_reverse.md)':
    '- [Array クラスの reverse インターフェイス](jp_array_reverse.md)',

    '- [Array class slice interface](array_slice.md)':
    '- [Array クラスの slice インターフェイス](jp_array_slice.md)',

    '- [Array class length interface](array_length.md)':
    '- [Array クラスの length (配列の長さ取得) のインターフェイス](jp_array_length.md)',

    '- [Array class join interface](array_join.md)':
    '- [Array クラスの join (値の連結文字列生成) のインターフェイス](jp_array_join.md)',

    '- [Array class index_of interface](array_index_of.md)':
    '- [Array クラスの index_of (値のインデックス取得) のインターフェイス](jp_array_index_of.md)',

    '- [Array class clear interface](array_clear.md)':
    '- [Array クラスの clear インターフェイス](jp_array_clear.md)',

    '- [Array class comparison interfaces](array_comparison.md)':
    '- [Array クラスの比較の各インターフェイス](jp_array_comparison.md)',

    '- [Dictionary class](dictionary.md)':
    '- [Dictionary クラス](jp_dictionary.md)',

    '- [Dictionary class generic type settings](dictionary_generic.md)':
    '- [Dictionary クラスのジェネリックの型設定](jp_dictionary_generic.md)',

    '- [Dictionary class get interface](dictionary_get.md)':
    '- [Dictionary クラスの get インターフェイス](jp_dictionary_get.md)',

    '- [Dictionary class length interface](dictionary_length.md)':
    '- [Dictionary クラスの length インターフェイス](jp_dictionary_length.md)',

    '- [Point2D class](point2d.md)':
    '- [Point2D クラス](jp_point2d.md)',

    '---':
    '---',

    '## DisplayObject and GraphicsBase classes':
    '## DisplayObject と GraphicsBase の各クラス',

    'The `DisplayObject` class is the base class for each display object. The `GraphicsBase` class is the `DisplayObject` subclass and the base class of each graphic class, such as the `Rectangle` class.':  # noqa
    '`DisplayObject`クラスは各表示用のオブジェクトの基底クラスです。`GraphicsBase`クラスは`DisplayObject`のサブクラスであり、且つ`Rectangle`などの各グラフィックスの基底クラスとなります。',  # noqa

    '### Table of contents':
    '### Table of contents',

    '- [DisplayObject class](display_object.md)':
    '- [DisplayObject クラス](jp_display_object.md)',

    '- [DisplayObject and GraphicsBase classes basic properties abstract](display_object_and_graphics_base_prop_abstract.md)':  # noqa
    '- [DisplayObject と GraphicsBase の各クラスの基本的な各属性の概要](jp_display_object_and_graphics_base_prop_abstract.md)',  # noqa

    '- [DisplayObject class x and y interfaces](display_object_x_and_y.md)':
    '- [DisplayObject クラスの x と y インターフェイス](jp_display_object_x_and_y.md)',

    '- [DisplayObject class parent interfaces](display_object_parent.md)':
    '- [DisplayObjectクラス parent （親要素属性）のインターフェイス](jp_display_object_parent.md)',  # noqa

    '- [DisplayObject class visible interface](display_object_visible.md)':
    '- [DisplayObject クラスの visible (表示・非表示) のインターフェイス](jp_display_object_visible.md)',  # noqa

    '- [DisplayObject class get_css and set_css interfaces](display_object_get_and_set_css.md)':  # noqa
    '- [DisplayObject クラスの get_css と set_css の各インターフェイス](jp_display_object_get_and_set_css.md)',  # noqa

    '- [DisplayObject class mouse event binding interfaces](display_object_mouse_event.md)':  # noqa
    '- [DisplayObject クラスのマウスイベント設定の各インターフェイス](jp_display_object_mouse_event.md)',  # noqa

    '- [GraphicsBase class rotation_around_center interface](graphics_base_rotation_around_center.md)':  # noqa
    '- [GraphicsBase クラスの rotation_around_center (中央座標基準の回転) インターフェイス](jp_graphics_base_rotation_around_center.md)',  # noqa

    '- [GraphicsBase class rotation_around_point interfaces](graphics_base_rotation_around_point.md)':  # noqa
    '- [GraphicsBase クラスの rotation_around_point (指定座標基準の回転) の各インターフェイス](jp_graphics_base_rotation_around_point.md)',  # noqa

    '- [GraphicsBase class scale_from_center interfaces](graphics_base_scale_from_center.md)':  # noqa
    '- [GraphicsBase クラスの scale_from_center (中央座標基準の拡縮) の各インターフェイス](jp_graphics_base_scale_from_center.md)',  # noqa

    '- [GraphicsBase class scale_from_point interfaces](graphics_base_scale_from_point.md)':  # noqa
    '- [GraphicsBase クラスの scale_from_point (指定座標基準の拡縮) の各インターフェイス](jp_graphics_base_scale_from_point.md)',  # noqa

    '- [GraphicsBase class flip_x and flip_y interfaces](graphics_base_flip_interfaces.md)':  # noqa
    '- [GraphicsBase クラスの flip_x (横軸の反転) と flip_y (縦軸の反転) のインターフェイス](jp_graphics_base_flip_interfaces.md)',  # noqa

    '- [GraphicsBase class skew_x and skew_y interfaces](graphics_base_skew.md)':  # noqa
    '- [GraphicsBase クラスの skew_x (X軸の歪み) と skew_y (Y軸の歪み) のインターフェイス](jp_graphics_base_skew.md)',  # noqa

    '---':
    '---',

    '## Graphics class':
    '## Graphics クラス',

    '### Table of contents':
    '### Table of contents',

    'The `Graphics` class handles each vector graphics drawing.':
    '`Graphics`クラスは各ベクターグラフィックスの描画処理を扱います。',

    '- [Draw interfaces abstract](draw_interfaces_abstract.md)':
    '- [描画の各インターフェイスの概要](jp_draw_interfaces_abstract.md)',

    '- [Graphics class](graphics.md)':
    '- [Graphics クラス](jp_graphics.md)',

    '- [Graphics class begin_fill interface](graphics_begin_fill.md)':
    '- [Graphics クラスの begin_fill (塗りの設定)のインターフェイス](jp_graphics_begin_fill.md)',  # noqa

    '- [Graphics class line_style interface](graphics_line_style.md)':
    '- [Graphics クラスの line_style (線のスタイル設定)のインターフェイス](jp_graphics_line_style.md)',  # noqa

    '- [Graphics class draw_rect interface](graphics_draw_rect.md)':
    '- [Graphics クラスの draw_rect (四角の描画)のインターフェイス](jp_graphics_draw_rect.md)',  # noqa

    '- [Graphics class draw_round_rect interface](graphics_draw_round_rect.md)':  # noqa
    '- [Graphics クラスの draw_round_rect (角丸の四角の描画)のインターフェイス](jp_graphics_draw_round_rect.md)',  # noqa

    '- [Graphics class draw_circle interface](graphics_draw_circle.md)':
    '- [Graphics クラスの draw_circle (円の描画)のインターフェイス](jp_graphics_draw_circle.md)',  # noqa

    '- [Graphics class draw_ellipse interface](graphics_draw_ellipse.md)':
    '- [Graphics クラスの draw_ellipse (楕円描画) のインターフェイス](jp_graphics_draw_ellipse.md)',  # noqa

    '- [Graphics class move_to and line_to interfaces](graphics_move_to_and_line_to.md)':  # noqa
    '- [Graphics クラスの move_to (線の描画位置の変更)と line_to (指定座標への線の描画)のインターフェイス](jp_graphics_move_to_and_line_to.md)',  # noqa

    '- [Graphics class draw_line interface](graphics_draw_line.md)':
    '- [Graphics クラスの draw_line (線の描画)のインターフェイス](jp_graphics_draw_line.md)',

    '- [Graphics class draw_dotted_line interface](graphics_draw_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_dotted_line (点線の描画)のインターフェイス](jp_graphics_draw_dotted_line.md)',  # noqa

    '- [Graphics class draw_dashed_line interface](graphics_draw_dashed_line.md)':  # noqa
    '- [Graphics クラスの draw_dashed_line (破線の描画)のインターフェイス](jp_graphics_draw_dashed_line.md)',  # noqa

    '- [Graphics class draw_round_dotted_line interface](graphics_draw_round_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_round_dotted_line (点線(丸)の描画)のインターフェイス](jp_graphics_draw_round_dotted_line.md)',  # noqa

    '- [Graphics class draw_dash_dotted_line interface](graphics_draw_dash_dotted_line.md)':  # noqa
    '- [Graphics クラスの draw_dash_dotted_line (一点鎖線の描画)のインターフェイス](jp_graphics_draw_dash_dotted_line.md)',  # noqa

    '- [Graphics class draw_polygon interface](graphics_draw_polygon.md)':
    '- [Graphics クラスの draw_polygon (多角形描画)のインターフェイス](jp_graphics_draw_polygon.md)',  # noqa

    '- [Graphics class fill_color interface](graphics_fill_color.md)':
    '- [Graphics クラスの fill_color (塗り設定)のインターフェイス](jp_graphics_fill_color.md)',  # noqa

    '- [Graphics class fill_alpha interface](graphics_fill_alpha.md)':
    '- [Graphics クラスの fill_alpha (塗りの透明度設定)のインターフェイス](jp_graphics_fill_alpha.md)',  # noqa

    '- [Graphics class line_color interface](graphics_line_color.md)':
    '- [Graphics クラスの line_color (線の色設定)のインターフェイス](jp_graphics_line_color.md)',  # noqa

    '- [Graphics class line_alpha interface](graphics_line_alpha.md)':
    '- [Graphics クラスの line_color (線の透明度設定)のインターフェイス](jp_graphics_line_alpha.md)',  # noqa

    '- [Graphics class line_thickness interface](graphics_line_thickness.md)':  # noqa
    '- [Graphics クラスの line_color (線幅設定)のインターフェイス](jp_graphics_line_thickness.md)',  # noqa

    '- [Graphics class line_dot_setting interface](graphics_line_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_dot_setting (点線設定)のインターフェイス](jp_graphics_line_dot_setting.md)',  # noqa

    '- [Graphics class line_dash_setting interface](graphics_line_dash_setting.md)':  # noqa
    '- [Graphics クラスの line_dash_setting (破線設定)のインターフェイス](jp_graphics_line_dash_setting.md)',  # noqa

    '- [Graphics class line_round_dot_setting interface](graphics_line_round_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_round_dot_setting (点線(丸)設定)のインターフェイス](jp_graphics_line_round_dot_setting.md)',  # noqa

    '- [Graphics class line_dash_dot_setting interface](graphics_line_dash_dot_setting.md)':  # noqa
    '- [Graphics クラスの line_dash_dot_setting (一点鎖線設定)のインターフェイス](jp_graphics_line_dash_dot_setting.md)',  # noqa

    '- [Graphics class clear interface](graphics_clear.md)':
    '- [Graphics クラスの clear インターフェイス](jp_graphics_clear.md)',

    '---':
    '---',

    '## Common event interfaces':
    '## イベントの共通の各インターフェイス',

    '### Table of contents':
    '### Table of contents',

    '- [About the handler options type](about_handler_options_type.md)':
    '- [ハンドラの options パラメーターの型について](jp_about_handler_options_type.md)',

    '- [Event class prevent_default and stop_propagation interfaces](event_prevent_default_and_stop_propagation.md)':  # noqa
    '- [Event クラスの prevent_default と stop_propagation の各インターフェイス](jp_event_prevent_default_and_stop_propagation.md)',  # noqa

    '- [bind_custom_event and trigger_custom_event interfaces](bind_and_trigger_custom_event.md)':  # noqa
    '- [bind_custom_event と trigger_custom_event の各インターフェイス](jp_bind_and_trigger_custom_event.md)',  # noqa

    '---':
    '---',

    '## MouseEvent class and mouse event binding':
    '## MouseEvent クラスとマウスイベントの設定',

    '### Table of contents':
    '### Table of contents',

    '- [MouseEvent interfaces abstract](mouse_event_abstract.md)':
    '- [MouseEvent クラスの各インターフェイスの概要](jp_mouse_event_abstract.md)',

    '- [Basic mouse event interfaces](mouse_event_basic.md)':
    '- [基本的なマウスイベントの各インターフェイス](jp_mouse_event_basic.md)',

    '- [click interface](click.md)':
    '- [click インターフェイス](jp_click.md)',

    '- [dblclick interface](dblclick.md)':
    '- [dblclick インターフェイス](jp_dblclick.md)',

    '- [mousedown and mouseup interfaces](mousedown_and_mouseup.md)':
    '- [mousedown と mouseup のインターフェイス](jp_mousedown_and_mouseup.md)',

    '- [mouseover and mouseout interfaces](mouseover_and_mouseout.md)':
    '- [mouseover と mouseout のインターフェイス](jp_mouseover_and_mouseout.md)',

    '- [mousemove interface](mousemove.md)':
    '- [mousemove インターフェイス](jp_mousemove.md)',

    '---':
    '---',

    '## Branch instruction':
    '## 条件分岐の制御',

    '### Table of contents':
    '### Table of contents',

    '- [If class](if.md)':
    '- [If クラス](jp_if.md)',

    '- [Elif class](elif.md)':
    '- [Elif クラス](jp_elif.md)',

    '- [Else class](else.md)':
    '- [Else クラス](jp_else.md)',

    '- [Each branch instruction class scope variables reverting setting](branch_instruction_variables_reverting_setting.md)':  # noqa
    '- [各条件分岐のクラスのスコープ内の変数値の復元設定](jp_branch_instruction_variables_reverting_setting.md)',  # noqa

    '- [Return class](return.md)':
    '- [Return クラス](jp_return.md)',

    '---':
    '---',

    '## Loop':
    '## ループ',

    '### Table of contents':
    '### Table of contents',

    '- [For loop class](for.md)':
    '- [ループ用の For クラス](jp_for.md)',

    '- [Continue class](continue.md)':
    '- [Continue クラス](jp_continue.md)',

    '---':
    '---',

    '## Timer':
    '## タイマー',

    '### Table of contents':
    '### Table of contents',

    '- [Timer class](timer.md)':
    '- [Timer クラス](jp_timer.md)',

    '- [TimerEvent class](timer_event.md)':
    '- [TimerEvent クラス](jp_timer_event.md)',

    '- [Timer class delay setting](timer_delay.md)':
    '- [Timer クラスの delay 設定](jp_timer_delay.md)',

    '- [FPS enum](fps.md)':
    '- [FPS の enum](jp_fps.md)',

    '- [Timer class repeat_count setting](timer_repeat_count.md)':
    '- [Timer クラスの repeat_count 設定](jp_timer_repeat_count.md)',

    '- [Timer class start and stop interfaces](timer_start_and_stop.md)':
    '- [Timer クラスの start と stop の各インターフェイス](jp_timer_start_and_stop.md)',

    '- [Timer class timer_complete interface](timer_complete.md)':
    '- [Timer クラスの timer_complete インターフェイス](jp_timer_complete.md)',

    '- [Timer class reset interface](timer_reset.md)':
    '- [Timer クラスの reset インターフェイス](jp_timer_reset.md)',

    '---':
    '---',

    '## Animation':
    '## アニメーション',

    '### Table of contents':
    '### Table of contents',

    '- [Animation interfaces abstract](animation_interfaces_abstract.md)':
    '- [アニメーションの各インターフェイスの概要](jp_animation_interfaces_abstract.md)',

    '- [AnimationEvent class](animation_event.md)':
    '- [AnimationEvent クラス](jp_animation_event.md)',

    '- [Animation duration setting](animation_duration.md)':
    '- [Animation クラスの duration 設定](jp_animation_duration.md)',

    '- [Animation delay setting](animation_delay.md)':
    '- [Animation クラスの delay 設定](jp_animation_delay.md)',

    '- [Each animation interface return value](animation_return_value.md)':
    '- [アニメーションの各インターフェイスの返却値について](jp_animation_return_value.md)',

    '- [AnimationBase class start interface](animation_base_start.md)':
    '- [AnimationBase クラスの start インターフェイス](jp_animation_base_start.md)',

    '- [AnimationBase class animation_complete interface](animation_complete.md)':  # noqa
    '- [AnimationBase クラスの animation_complete インターフェイス](jp_animation_complete.md)',  # noqa

    '- [AnimationBase class interfaces method chaining](animation_method_chaining.md)':  # noqa
    '- [AnimationBase クラスの各インターフェイスのメソッドチェーンについて](jp_animation_method_chaining.md)',  # noqa

    '- [AnimationBase class target property](animation_base_target.md)':
    '- [AnimationBase クラスの target 属性](jp_animation_base_target.md)',

    '- [Animation pause and play interfaces](animation_pause_and_play.md)':
    '- [アニメーションの pause と play の各インターフェイス](jp_animation_pause_and_play.md)',

    '- [Animation reset interface](animation_reset.md)':
    '- [アニメーションの reset インターフェイス](jp_animation_reset.md)',

    '- [Animation finish interface](animation_finish.md)':
    '- [アニメーションの finish インターフェイス](jp_animation_finish.md)',

    '- [Animation reverse interface](animation_reverse.md)':
    '- [アニメーションの reverse インターフェイス](jp_animation_reverse.md)',

    '- [animation_time interface](animation_time.md)':
    '- [animation_time インターフェイス](jp_animation_time.md)',

    '- [Easing enum](easing_enum.md)':
    '- [イージングの enum](jp_easing_enum.md)',

    '- [Sequential animation setting](sequential_animation.md)':
    '- [連続したアニメーション設定](jp_sequential_animation.md)',

    '- [animation_parallel interface](animation_parallel.md)':
    '- [animation_parallel インターフェイス](jp_animation_parallel.md)',

    '- [animation_move interface](animation_move.md)':
    '- [animation_move インターフェイス](jp_animation_move.md)',

    '- [animation_x interface](animation_x.md)':
    '- [animation_x インターフェイス](jp_animation_x.md)',

    '- [animation_y interface](animation_y.md)':
    '- [animation_y インターフェイス](jp_animation_y.md)',

    '- [animation_width and animation_height interfaces](animation_width_and_height.md)':  # noqa
    '- [animation_width と animation_height の各インターフェイス](jp_animation_width_and_height.md)',  # noqa

    '- [animation_fill_color interface](animation_fill_color.md)':
    '- [animation_fill_color インターフェイス](jp_animation_fill_color.md)',

    '- [animation_fill_alpha interface](animation_fill_alpha.md)':
    '- [animation_fill_alpha インターフェイス](jp_animation_fill_alpha.md)',

    '- [animation_line_color interface](animation_line_color.md)':
    '- [animation_line_color インターフェイス](jp_animation_line_color.md)',

    '- [animation_line_alpha interface](animation_line_alpha.md)':
    '- [animation_line_alpha インターフェイス](jp_animation_line_alpha.md)',

    '- [animation_line_thickness interface](animation_line_thickness.md)':
    '- [animation_line_thickness インターフェイス](jp_animation_line_thickness.md)',

    '- [animation_radius interface](animation_radius.md)':
    '- [animation_radius インターフェイス](jp_animation_radius.md)',

    '- [animation_rotation_around_center interface](animation_rotation_around_center.md)':  # noqa
    '- [animation_rotation_around_center インターフェイス](jp_animation_rotation_around_center.md)',  # noqa

    '- [animation_rotation_around_point interface](animation_rotation_around_point.md)':  # noqa
    '- [animation_rotation_around_point インターフェイス](jp_animation_rotation_around_point.md)',  # noqa

    '- [animation_scale_x_from_center and animation_scale_y_from_center interfaces](animation_scale_x_and_y_from_center.md)':  # noqa
    '- [animation_scale_x_from_center と animation_scale_y_from_center の各インターフェイス](jp_animation_scale_x_and_y_from_center.md)',  # noqa

    '- [animation_scale_x_from_point and animation_scale_y_from_point interfaces](animation_scale_x_and_y_from_point.md)':  # noqa
    '- [animation_scale_x_from_point と animation_scale_y_from_point の各インターフェイス](jp_animation_scale_x_and_y_from_point.md)',  # noqa

    '- [animation_skew_x interface](animation_skew_x.md)':
    '- [animation_skew_x インターフェイス](jp_animation_skew_x.md)',

    '---':
    '---',

    '## Other manipulation interfaces':
    '## その他の操作関係の各インターフェイス',

    '### Table of contents':
    '### Table of contents',

    '- [delete interface](delete.md)':
    '- [delete インターフェイス](jp_delete.md)',

    '---':
    '---',

    '## Debugging':
    '## デバッグ',

    '### Table of contents':
    '### Table of contents',

    '- [trace function interface](trace.md)':
    '- [trace 関数のインターフェイス](jp_trace.md)',

    '- [set_debug_mode interface](set_debug_mode.md)':
    '- [set_debug_mode インターフェイス](jp_set_debug_mode.md)',

    '- [unset_debug_mode interface](unset_debug_mode.md)':
    '- [unset_debug_mode インターフェイス](jp_unset_debug_mode.md)',

    '- [add_debug_info_setting decorator interface](add_debug_info_setting.md)':  # noqa
    '- [add_debug_info_setting のデコレーターのインターフェイス](jp_add_debug_info_setting.md)',  # noqa

    '---':
    '---',

    '## Testing':
    '## テスト',

    '### Table of contents':
    '### Table of contents',

    '- [JavaScript assertion interface basic behavior](assertion_basic_behavior.md)':  # noqa
    '- [JavaScriptの各アサーションのインターフェイスの基本的な挙動](jp_assertion_basic_behavior.md)',  # noqa

    '- [assert_equal and assert_not_equal interfaces](assert_equal_and_not_equal.md)':  # noqa
    '- [assert_equal と assert_not_equal の各インターフェイス](jp_assert_equal_and_not_equal.md)',  # noqa

    '- [assert_true and assert_false interfaces](assert_true_and_false.md)':
    '- [assert_true と assert_false の各インターフェイス](jp_assert_true_and_false.md)',  # noqa

    '- [assert_arrays_equal and assert_arrays_not_equal interfaces](assert_arrays_equal_and_arrays_not_equal.md)':  # noqa
    '- [assert_arrays_equal と assert_arrays_not_equal の各インターフェイス](jp_assert_arrays_equal_and_arrays_not_equal.md)',  # noqa

    '- [assert_dicts_equal and assert_dicts_not_equal interfaces](assert_dicts_equal_and_dicts_not_equal.md)':  # noqa
    '- [assert_dicts_equal と assert_dicts_not_equal の各インターフェイス](jp_assert_dicts_equal_and_dicts_not_equal.md)',  # noqa

    '- [assert_defined and assert_undefined interfaces](assert_defined_and_undefined.md)':  # noqa
    '- [assert_defined と assert_undefined の各インターフェイス](jp_assert_defined_and_undefined.md)',  # noqa

}



"""This module is for the translation mapping data of the
following document:

Document file: remove_children.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# remove_children interface':
    '# remove_children インターフェイス',

    'This page explains the `remove_children` method interface of container classes.':  # noqa
    'このページではコンテナの各クラスの`remove_children`メソッドのインターフェイスについて説明します。',

    '## What interface is this?':
    '## インターフェイス概要',

    'The `remove_children` method removes all children from a container instance.':  # noqa
    '`remove_children`メソッドはコンテナのインスタンスから全ての子のインスタンスを取り除きます。',

    '## Basic usage':
    '## 基本的な使い方',

    'The `remove_children` method takes no arguments to use.':
    '`remove_children`メソッドは引数の指定を必要としません。',

    'In the following example, if you click any rectangle, the handler calls the `remove_children` method and removes all children:':  # noqa
    '以下の例では、いずれかの四角をクリックした場合`remove_children`メソッドが呼ばれ全ての子が取り除かれます:',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The click event handler.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent[ap.Sprite]\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    sprite.remove_children()\n\n\nap.Stage(\n    stage_width=250, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'remove_children_basic_usage/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The click event handler.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent[ap.Sprite]\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    sprite.remove_children()\n\n\nap.Stage(\n    stage_width=250, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'remove_children_basic_usage/\')\n```',  # noqa

}



"""Typical callable utility implementations.
"""

import inspect
from inspect import Parameter
from inspect import Signature
from typing import Any
from typing import Callable
from typing import Dict

from apysc._html.debug_mode import add_debug_info_setting

empty = inspect.Signature.empty


@add_debug_info_setting(
    module_name=__name__)
def get_func_default_vals(*, func: Callable) -> Dict[str, Any]:
    """
    Get specified function's arguments default values.

    Parameters
    ----------
    func : Callable
        Target function (or method).

    Returns
    -------
    default_vals : dict
        Dictionary with an argument name at key and
        default value at value. This interface sets the
        `empty` object if an argument has no default value.
    """
    signature: Signature = inspect.signature(func)
    default_vals: Dict[str, Any] = {}
    parameter: Parameter
    for arg_name, parameter in signature.parameters.items():
        default_vals[arg_name] = parameter.default
    return default_vals


@add_debug_info_setting(module_name=__name__)
def get_arg_name_at(*, func: Callable, index: int) -> str:
    """
    Get specified index argument name from function.

    Parameters
    ----------
    func : Callable
        Target function (or method).
    index : int
        Argument index (start from 0).

    Returns
    -------
    arg_name : str
        Target index's argument name.
    """
    signature: Signature = inspect.signature(func)
    arg_name: str = ''
    for i, target_idx_arg_name in enumerate(signature.parameters.keys()):
        if i != index:
            continue
        arg_name = target_idx_arg_name
        break
    return arg_name


@add_debug_info_setting(module_name=__name__)
def get_name_and_arg_value_dict_from_args(
        *, func: Callable, args: list, kwargs: dict) -> Dict[str, Any]:
    """
    Get a dictionary with an argument name at key and
    specified argument values at value.

    Parameters
    ----------
    func : Callable
        Target function (or method).
    args : list
        Specified positional arguments.
    kwargs : dict
        Specified keyword arguments.

    Returns
    -------
    name_and_arg_value_dict : dict
        Dictionary that has an argument name at key and specified
        argument values at value.
    """
    name_and_arg_value_dict: Dict[str, Any] = {}
    for i, arg_value in enumerate(args):
        arg_name: str = get_arg_name_at(func=func, index=i)
        name_and_arg_value_dict[arg_name] = arg_value
    name_and_arg_value_dict.update(kwargs)
    return name_and_arg_value_dict


@add_debug_info_setting(module_name=__name__)
def get_method_class_name(*, method: Callable) -> str:
    """
    Get a specified method's class name.

    Parameters
    ----------
    method : Callable
        Target method.

    Returns
    -------
    class_name : str
        Target method's class name. This interface
        returns a blank string if a specified argument
        is not a method (e.g., function).
    """
    if not inspect.ismethod(method):
        return ''
    class_name: str = ''
    for class_ in inspect.getmro(type(method.__self__)):  # type: ignore
        if method.__name__ not in class_.__dict__:
            continue
        class_name = class_.__name__
    return class_name



"""This module is for the translation mapping data of the
following document:

Document file: array_join.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class join interface':
    '# Array クラスの join インターフェイス',

    'This page explains the `Array` class `join` method interface.':
    'このページでは`Array`クラスの`join`メソッドのインターフェイスについて説明します。',

    '## What interface is this?':
    '## インターフェイス概要',

    'The `join` method returns a joined `String` with the specified separator string.':  # noqa
    '`join`メソッドは引数に指定された区切り文字で連結された配列の値の`String`クラスの文字列を返却します。',

    '## Basic usage':
    '## 基本的な使い方',

    'The `join` method requires the `sep` argument as the separator, as follows:':  # noqa
    '`join`メソッドは以下のコード例のように区切り文字としての`sep`引数が必要になります。',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3])\njoined: ap.String = arr.join(sep=\',\')\nassert joined == \'1,2,3\'\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3])\njoined: ap.String = arr.join(sep=\',\')\nassert joined == \'1,2,3\'\n```',  # noqa

    '## join API':
    '## join API',

    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>':  # noqa
    '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa

    '**[Interface summary]** Join this array value with a specified separator string.<hr>':  # noqa
    '**[インターフェイス概要]** 指定された区切り文字を使ってこの配列の値を連結した文字列を生成します。<hr>',

    '**[Parameters]**':
    '**[引数]**',

    '- `sep`: String or str':
    '- `sep`: String or str',

    '  - Separator string.':
    '  - 区切り文字。',

    '<hr>':
    '<hr>',

    '**[Returns]**':
    '**[返却値]**',

    '- `joined`: String':
    '- `joined`: String',

    '  - Joined string.':
    '  - 連結された文字列。',

    '<hr>':
    '<hr>',

    '**[Examples]**':
    '**[コードサンプル]**',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 2, 3])\n>>> arr.join(sep=\', \')\nString(\'1, 2, 3\')\n```':  # noqa
    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 2, 3])\n>>> arr.join(sep=\', \')\nString(\'1, 2, 3\')\n```',  # noqa

}



"""This module is for the argument validations' decorators.

Mainly the following decorators exist.

- not_empty_string
    - Set the validation to check that a specified argument's string
        is not empty.
- handler_args_num
    - Set the validation to check a specified handler argument's
        number.
- handler_options_type
    - Set the validation to check a specified handler-options
        argument's type.
- is_event
    - Set the validation to check a specified argument's type
        is the `ap.Event` or its subclass type.
- is_num
    - Set the validation to check a specified argument's type
        is the number-related type.
- is_apysc_num
    - Set the validation to check a specified argument's type
        is the `ap.Int` or `ap.Number` type.
- is_integer
    - Set the validation to check a specified argument's type
        is the `int` or `ap.Int`.
- is_builtin_integer
    - Set the validation to check a specified argument's type
        is the built-in `int`.
- is_apysc_integer
    - Set the validation to check a specified argument's type
        is the `ap.Int`.
- num_is_gt_zero
    - Set the validation to check that a specified argument's value
        is greater than zero.
- num_is_gte_zero
    - Set the validation to check that a specified argument's value
        is greater than or equal to zero.
- num_is_0_to_1_range
    - Set the validation to check that a specified argument's value
        is 0.0 to 1.0 range.
- is_boolean
    - Set the validation to check that a specified argument's type
        is the `bool` or `ap.Boolean`.
- is_builtin_boolean
    - Set the validation to check that a specified argument's type
        is the built-in `bool`.
- is_apysc_boolean
    - Set the validation to check that a specified argument's type
        is the `ap.Boolean`.
- is_easing
    - Set the validation to check a specified argument's type
        is the `ap.Easing`.
- is_string
    - Set the validation to check a specified argument's type
        is the str or `ap.String`.
- is_builtin_string
    - Set the validation to check a specified argument's type
        is the Python built-in's `str`.
- is_hex_color_code_format
    - Set the validation to check a specified argument's value
        is a hexadecimal color code format.
- is_animations
    - Set the validation to check a specified argument's type
        is the list of `ap.AnimationBase`.
- is_vars_dict
    - Set the validation to check a specified argument's value
        is a variables' dictionary.
- is_display_object
    - Set the validation to check a specified argument's type
        is the `ap.DisplayObject` or its subclass type.
- is_display_object_container
    - Set the validation to check a specified argument's type
        is a container of a display object instance.
- is_point_2d
    - Set the validation to check a specified argument's type
        is the `ap.Point2D`.
- is_point_2ds
    - Set the validation to check a specified argument's type
        is the list of `ap.Point2D`.
- is_path_data_list
    - Set the validation to check a specified argument's type
        is the list of `ap.PathDataBase`.
- is_line_cap
    - Set the validation to check a specified argument's type
        is a line cap-related type.
- is_line_joints
    - Set the validation to check a specified argument's type
        is a line joints-related type.
- multiple_line_settings_are_not_set
    - Set the validation to check a specified argument's instance
        does not have multiple line settings.
- is_line_dot_setting
    - Set the validation to check a specified argument's type
        is the `ap.LineDotSetting`.
- is_line_dash_setting
    - Set the validation to check a specified argument's type
        is the `ap.LineDashSetting`.
- is_line_dash_dot_setting
    - Set the validation to check a specified argument's type
        is the `ap.LineDashDotSetting`.
- is_line_round_dot_setting
    - Set the validation to check a specified argument's type
        is the `ap.LineRoundDotSetting`.
- is_variable_name_interface_type
    - Set the validation to check a specified argument's type
        is the `ap.VariableNameInterface` or its subclass type.
"""

import functools
import inspect
from inspect import Signature
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import TypeVar

# pyright: reportInvalidTypeVarUse=false
_F = TypeVar('_F', bound=Callable)


def _extract_arg_value(
        *, args: Any, kwargs: Dict[str, Any],
        arg_position_index: int,
        callable_: Callable) -> Any:
    """
    Extract an argument value from a specified arguments' dictionary
    or list.

    Parameters
    ----------
    args : List[Any]
        A specified positional arguments' list.
    kwargs : Dict[str, Any]
        A specified keyword arguments' dictionary.'
    arg_position_index : int
        A target argument position index.
    callable_ : Callable
        A target function or method.

    Returns
    -------
    value : Any
        An extracted any value.
    """
    value: Any = None
    arg_name: str = _get_arg_name_by_index(
        callable_=callable_, arg_position_index=arg_position_index)
    default_val: Any = _get_default_val_by_arg_name(
        callable_=callable_, arg_name=arg_name)
    if arg_name in kwargs:
        value = kwargs[arg_name]
    elif len(args) - 1 >= arg_position_index:
        value = args[arg_position_index]
    if value is None and default_val is not None:
        return default_val
    return value


def _get_arg_name_by_index(
        *, callable_: Callable, arg_position_index: int) -> str:
    """
    Get an argument name from a specified argument position index.

    Parameters
    ----------
    callable_ : Callable
        A target function or method.
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    arg_name : str
        A target argument name.
    """
    signature: Signature = inspect.signature(callable_)
    if len(signature.parameters) - 1 < arg_position_index:
        raise IndexError(
            'A specified function has no argument parameter '
            f'at the index of {arg_position_index}'
            f'\nActual argument length: {len(signature.parameters)}')
    arg_name: str = ''
    for i, (arg_name_, _) in enumerate(signature.parameters.items()):
        if i == arg_position_index:
            arg_name = arg_name_
    return arg_name


def _get_callable_and_arg_names_msg(
        *, callable_: Callable, arg_position_index: int) -> str:
    """
    Get a function or method and argument names' message
    for an additional error message.

    Parameters
    ----------
    callable_ : Callable
        A target function or method.
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    callable_and_arg_names_msg : str
        A function or method and argument names' message.
    """
    arg_name: str = _get_arg_name_by_index(
        callable_=callable_, arg_position_index=arg_position_index)
    callable_and_arg_names_msg: str = (
        f'Target callable name: {callable_.__name__}'
        f'\nTarget argument name: {arg_name}'
    )
    return callable_and_arg_names_msg


def _get_default_val_by_arg_name(
        *, callable_: Callable, arg_name: str) -> Any:
    """
    Get a default value of a given name's argument.

    Parameters
    ----------
    callable_ : Callable
        A target function or method.
    arg_name : str
        A target argument name.

    Returns
    -------
    default_val : Any
        A default value of a given name's argument.
    """
    default_val: Any = None
    signature: Signature = inspect.signature(callable_)
    for target_arg_name, parameter in signature.parameters.items():
        if target_arg_name != arg_name:
            continue
        default_val = parameter.default
        break
    if default_val == inspect.Signature.empty:
        return None
    return default_val


def not_empty_string(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's string
    is not empty.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.string_validation import \
                validate_not_empty_string
            string: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_not_empty_string(
                string=string,
                additional_err_msg=(
                    'An argument\'s string value must not be empty.'
                    f'\n{callable_and_arg_names_msg}'
                ))

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def handler_args_num(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified handler argument's
    number.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.handler_validation import \
                validate_handler_args_num
            handler: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_handler_args_num(
                handler=handler,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def handler_options_type(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified handler-options
    argument's type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.handler_validation import \
                validate_options_type
            options: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_options_type(
                options=options,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_event(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.Event` or its subclass type.

    Parameters
    ----------
    arg_position_index : int
        _description_

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.event_validation import validate_event
            event: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_event(
                e=event, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_num(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the number-related type.

    Parameters
    ----------
    arg_position_index : int
        _description_

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import validate_num
            num: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_num(
                num=num,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_apysc_num(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.Int` or `ap.Number` type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._type.number_value_interface import NumberValueInterface
            num: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(num, NumberValueInterface):
                raise TypeError(
                    'A specified argument value is not the `ap.Int` or '
                    f'`ap.Number` type: {type(num)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_integer(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `int` or `ap.Int`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import validate_integer
            integer: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_integer(
                integer=integer,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_builtin_integer(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the built-in `int`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import \
                validate_builtin_integer
            integer: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_builtin_integer(
                integer=integer,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_apysc_integer(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.Int`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            integer: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(integer, ap.Int):
                raise TypeError(
                    'A specified argument value is not the `ap.Int` '
                    f'type: {type(integer)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def num_is_gt_zero(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's value
    is greater than zero.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import \
                validate_num_is_gt_zero
            num: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_num_is_gt_zero(
                num=num, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def num_is_gte_zero(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's value
    is greater than or equal to zero.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import \
                validate_num_is_gte_zero
            num: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_num_is_gte_zero(
                num=num, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def num_is_0_to_1_range(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's value
    is 0.0 to 1.0 range.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.number_validation import \
                validate_num_is_0_to_1_range
            num: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_num_is_0_to_1_range(
                num=num, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_boolean(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's type
    is the `bool` or `ap.Boolean`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.bool_validation import validate_bool
            boolean: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_bool(
                value=boolean, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_builtin_boolean(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's type
    is the built-in `bool`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.bool_validation import validate_builtin_bool
            boolean: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_builtin_bool(
                value=boolean, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_apysc_boolean(*, arg_position_index: int) -> _F:
    """
    Set the validation to check that a specified argument's type
    is the `ap.Boolean`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            boolean: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(boolean, ap.Boolean):
                raise TypeError(
                    'A specified argument value is not a `Boolean` type: '
                    f'{type(boolean)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_easing(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.Easing`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            easing: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(easing, ap.Easing):
                raise TypeError(
                    'A specified easing argument\'s type is not the '
                    f'ap.Easing: {type(easing)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_string(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the str or `ap.String`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.string_validation import \
                validate_string_type
            string: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_string_type(
                string=string, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_builtin_string(*, arg_position_index: int, optional: bool) -> _F:
    """
    Set the validation to check a specified argument's type
    is the Python built-in's `str`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.
    optional : bool, optional
        A boolean indicating whether a target argument accepts
        optional None value or not.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.string_validation import \
                validate_builtin_string_type
            string: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not optional:
                validate_builtin_string_type(
                    string=string,
                    additional_err_msg=callable_and_arg_names_msg)
            elif string is not None:
                validate_builtin_string_type(
                    string=string,
                    additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_hex_color_code_format(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's value
    in a hexadecimal color code format.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._color import color_util
            from apysc._validation.color_validation import \
                validate_hex_color_code_format
            hex_color_code: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            hex_color_code = color_util.remove_color_code_sharp_symbol(
                hex_color_code=hex_color_code)
            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_hex_color_code_format(
                hex_color_code=hex_color_code,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_animations(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the list of `ap.AnimationBase`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            animations: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)
            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)

            if not isinstance(animations, list):
                raise TypeError(
                    'A specified animations list must be a list: '
                    f'{type(animations)}'
                    f'\n{callable_and_arg_names_msg}')

            for i, animation in enumerate(animations):
                if isinstance(animation, ap.AnimationBase):
                    continue
                raise TypeError(
                    'A specified animations\' list cannot contain '
                    f'non-animation instance: {type(animation)}'
                    f'Invalid index: {i}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_vars_dict(
        *, arg_position_index: int, optional: bool = True) -> _F:
    """
    Set the validation to check a specified argument's value
    is a variables' dictionary.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.
    optional : bool, optional
        A boolean indicating whether a target argument accepts
        optional None value or not.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            vars_dict: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)
            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)

            if optional and not (
                isinstance(
                    vars_dict,
                    dict) or vars_dict is None):
                raise TypeError(
                    'A specified variables argument value is not a dictionary '
                    f'or None: {type(vars_dict)}'
                    f'\nDictionary value: {vars_dict}'
                    f'\n{callable_and_arg_names_msg}')

            if not optional and not isinstance(vars_dict, dict):
                raise TypeError(
                    'A specified variables argument value is not '
                    f'a dictionary: {vars_dict}'
                    f'\n{callable_and_arg_names_msg}')

            if isinstance(vars_dict, dict):
                for key in vars_dict.keys():
                    if isinstance(key, str):
                        continue
                    raise ValueError(
                        'A specified variables argument dictionary\'s '
                        f'key cannot contain a non-str value: {type(key)}'
                        f', {key}'
                        f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_display_object(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.DisplayObject` or its subclass type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.display_validation import \
                validate_display_object
            display_object: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_display_object(
                display_object=display_object,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_display_object_container(
        *, arg_position_index: int,
        optional: bool) -> _F:
    """
    Set the validation to check a specified argument's type
    is a container of a display object instance.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.
    optional : bool, optional
        A boolean indicating whether a target argument accepts
        optional None value or not.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.display_validation import \
                validate_display_object_container
            container_object: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not optional:
                validate_display_object_container(
                    container_object=container_object,
                    additional_err_msg=callable_and_arg_names_msg)
            elif container_object is not None:
                validate_display_object_container(
                    container_object=container_object,
                    additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_point_2d(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.Point2D`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """
    from apysc._validation.geom_validation import validate_point_2d_type

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            point: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_point_2d_type(
                point=point, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_point_2ds(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the list of `ap.Point`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """
    from apysc._validation.geom_validation import validate_point_2d_type

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            points: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)
            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if (
                    not isinstance(points, list)
                    and not isinstance(points, ap.Array)):
                raise TypeError(
                    'A specified points argument type is not the list or '
                    f'ap.Array: {type(points)}'
                    f'\n{callable_and_arg_names_msg}')

            value: List[ap.Point2D] = []
            if isinstance(points, list):
                value = points
            elif isinstance(points, ap.Array):
                value = points._value
            for point in value:
                validate_point_2d_type(
                    point=point,
                    additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_path_data_list(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the list of `ap.PathDataBase`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            from apysc._validation.path_validation import \
                validate_path_data_list
            path_data_list = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)

            if not isinstance(path_data_list, list):
                raise TypeError(
                    'A specified `path_data_list`\'s type is not the '
                    f'list: {type(path_data_list)}'
                    f'\n{callable_and_arg_names_msg}')

            for path_data in path_data_list:
                if isinstance(path_data, ap.PathDataBase):
                    continue
                raise TypeError(
                    'A value of path_data_list argument is not a '
                    'type of `ap.PathDataBase`\'s subclass: '
                    f'{type(path_data)}')
            validate_path_data_list(path_data_list=path_data_list)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_cap(
        *, arg_position_index: int,
        optional: bool) -> _F:
    """
    Set the validation to check a specified argument's type
    is a line cap-related type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.
    optional : bool
        A boolean indicating whether a specified argument can be
        the `None`.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.display_validation import validate_line_cap
            cap: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not optional:
                validate_line_cap(
                    cap=cap, additional_err_msg=callable_and_arg_names_msg)
            elif cap is not None:
                validate_line_cap(
                    cap=cap, additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_joints(
        *, arg_position_index: int,
        optional: bool) -> _F:
    """
    Set the validation to check a specified argument's type
    is a line joints-related type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.
    optional : bool
        A boolean indicating whether a specified argument can be
        the `None`.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.display_validation import \
                validate_line_joints
            joints: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not optional:
                validate_line_joints(
                    joints=joints,
                    additional_err_msg=callable_and_arg_names_msg)
            elif joints is not None:
                validate_line_joints(
                    joints=joints,
                    additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def multiple_line_settings_are_not_set(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's instance
    does not have multiple line settings.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.display_validation import \
                validate_multiple_line_settings_are_not_set
            instance: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            result: Any = callable_(*args, **kwargs)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_multiple_line_settings_are_not_set(
                any_instance=instance,
                additional_err_msg=callable_and_arg_names_msg)

            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_dot_setting(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.LineDotSetting`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """
    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            setting: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(setting, (type(None), ap.LineDotSetting)):
                raise TypeError(
                    'A specified setting is not the `ap.LineDotSetting` '
                    f'type or None: {type(setting)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_dash_setting(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.LineDashSetting`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            setting: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(setting, (type(None), ap.LineDashSetting)):
                raise TypeError(
                    'A specified setting is not the `ap.LineDashSetting`'
                    f' type or None: {type(setting)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_dash_dot_setting(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.LineDashDotSetting`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            setting: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(
                    setting, (type(None), ap.LineDashDotSetting)):
                raise TypeError(
                    'A specified setting is not the '
                    '`ap.LineDashDotSetting` type or None: '
                    f'{type(setting)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_line_round_dot_setting(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.LineRoundDotSetting`.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            import apysc as ap
            setting: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            if not isinstance(setting, (type(None), ap.LineRoundDotSetting)):
                raise TypeError(
                    'A specified setting is not the `ap.LineRoundDotSetting` '
                    f'type or None: {type(setting)}'
                    f'\n{callable_and_arg_names_msg}')

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore


def is_variable_name_interface_type(*, arg_position_index: int) -> _F:
    """
    Set the validation to check a specified argument's type
    is the `ap.VariableNameInterface` or its subclass type.

    Parameters
    ----------
    arg_position_index : int
        A target argument position index.

    Returns
    -------
    wrapped : Callable
        Wrapped callable object.
    """

    def wrapped(callable_: _F) -> _F:

        @functools.wraps(callable_)
        def inner_wrapped(*args: Any, **kwargs: Any) -> Any:
            from apysc._validation.variable_name_validation import \
                validate_variable_name_interface_type
            instance: Any = _extract_arg_value(
                args=args, kwargs=kwargs,
                arg_position_index=arg_position_index, callable_=callable_)

            callable_and_arg_names_msg: str = _get_callable_and_arg_names_msg(
                callable_=callable_, arg_position_index=arg_position_index)
            validate_variable_name_interface_type(
                instance=instance,
                additional_err_msg=callable_and_arg_names_msg)

            result: Any = callable_(*args, **kwargs)
            return result

        return inner_wrapped  # type: ignore

    return wrapped  # type: ignore



"""This module is for the translation mapping data of the
following document:

Document file: graphics_clear.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics clear interface':
    '# Graphics クラスの clear インターフェイス',

    'This page explains the `Graphics` class `clear` method interface.':
    'このページでは`Graphics`クラスの`clear`メソッドのインターフェイスについて説明します。',

    '## What interface is this?':
    '## インターフェイス概要',

    'The `clear` method removes all graphics and clears (resets) fill and line settings.':  # noqa
    '`clear`メソッドは全ての描画済みグラフィックスを取り除き、塗りと線の設定をリセットします。',

    '## Basic usage':
    '## 基本的な使い方',

    'The `clear` method takes no arguments.':
    '`clear`メソッドは引数の指定を必要としません。',

    'The handler calls the `clear` method in the following example if you click any rectangle.':  # noqa
    '以下の例では四角をクリックした際にハンドラ内で`clear`メソッドが呼ばれます:',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The click event handler.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent[ap.Sprite]\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    ap.assert_equal(sprite.graphics.fill_color, \'#00aaff\')\n    sprite.graphics.clear()\n    ap.assert_equal(sprite.graphics.fill_color, \'\')\n\n\nap.Stage(\n    stage_width=250, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_clear_basic_usage/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The click event handler.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent[ap.Sprite]\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    ap.assert_equal(sprite.graphics.fill_color, \'#00aaff\')\n    sprite.graphics.clear()\n    ap.assert_equal(sprite.graphics.fill_color, \'\')\n\n\nap.Stage(\n    stage_width=250, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_clear_basic_usage/\')\n```',  # noqa

}
